﻿namespace KPGBase.Excercises
{
    internal class LineSegment
    {
        public PointF A;
        public PointF B;
    }

    internal class Excercise08 : BasicExcercise
    {
        public Excercise08(Bitmap canvas) : base(canvas) { }
        
        //Koch snowflake recursive surve segments
        private LineSegment[] currentGen;   // currently displayed generated recursive curve
        private LineSegment[] nextGen;      // next generation of recursive curve
        private PointF translation;         // translation vector so the starting shape is in the center

        // ID of displayed curve, 1 = koch, 2 = special koch, 3 = dragon
        private int curveID = 1;

        internal void ResetGeneration()
        {
            currentGen = null;
            nextGen = null;

            ClearImage();
        }

        /// <summary>
        /// Generate curve and redraw
        /// </summary>
        internal void GenerateAndRedraw()
        {
            // first iteration - need to init
            if (currentGen == null) 
            {
                // Koch curve
                if (curveID == 1)
                    currentGen = FirstGeneration();
                // Bothsided Koch curve
                else if (curveID == 2)
                    currentGen = FirstGenerationSpecialKoch();
                // Dragon curve
                else if (curveID == 3)
                    currentGen = FirstGenerationDragon();
            }
            // next iteration
            else
            {
                // Koch curve
                if (curveID == 1)
                    currentGen = NextKochGeneration(currentGen);
                // Special Koch
                else if (curveID == 2)
                    currentGen = NextSpecialKochGeneration(currentGen);
                // Dragon curve
                else if (curveID == 3)
                    currentGen = NextDragonGeneration(currentGen);
            }

            // Reset transform and apply base translation
            g.ResetTransform();
            g.TranslateTransform(translation.X, translation.Y);

            // Draw fractal
            DrawSegments(currentGen);
        }

        /// <summary>
        /// Draw the line segments of a fractal
        /// </summary>
        /// <param name="coordinates"> Line segments to draw </param>
        private void DrawSegments(LineSegment[]? coordinates)
        {
            ClearImage();
            Pen pen0 = new Pen(Color.White, 2);

            if (coordinates == null)
                return;

            for (int i = 0; i < coordinates.Length; i++)
                g.DrawLine(pen0, coordinates[i].A, coordinates[i].B);
        }


#region Koch snowflake

        /// <summary>
        /// Create first generation for koch and dragon curve
        /// - an equilateral triangle
        /// </summary>
        /// <returns> Array with the sides of the triangle as line segments </returns>
        internal LineSegment[] FirstGeneration()
        {
            LineSegment[] triangle = new LineSegment[3];
            for (int i = 0; i < triangle.Length; i++)
                triangle[i] = new LineSegment();

            // Create an equilateral triangle
            // - create vector of one side s=(x, 0)
            // - first side will be from (0,0) to (x, 0)
            // - rotate the vector s by 60 degrees using the prepared method RotateVector(PointF, radians)
            // - create second and third side

            // Set up attribute translation
            // - value of translation to the center of the canvas
            // - centers the created triangle in the canvas

            float lineLength = 0.7f * canvas.Height;
            PointF lineShape = new PointF(lineLength, 0);
            double sixtyDegrees = Math.PI / 3;

            triangle[0].A = new PointF(0, 0);
            triangle[0].B = new PointF(lineShape.X, lineShape.Y);

            lineShape = RotateVector(lineShape, -sixtyDegrees);

            triangle[1].A = triangle[0].B;
            triangle[1].B = lineShape;
            triangle[2].A = lineShape;
            triangle[2].B = triangle[0].A;

            // Compute translation vector
            float x = 0.5f * (canvas.Width - lineLength);
            float y = Math.Abs(lineShape.Y) + 0.5f * (canvas.Height - 1.25f * Math.Abs(lineShape.Y));
            translation = new PointF(x, y);

            return triangle;
        }

        /// <summary>
        /// Create the next iteration of koch snowflake
        /// </summary>
        /// <param name="generation"> Current generation </param>
        /// <returns> Next generation </returns>
        private LineSegment[] NextKochGeneration(LineSegment[] generation)
        {
            // For each side 4 new will be created
            nextGen = new LineSegment[generation.Length * 4];
            for (int i = 0; i < nextGen.Length; i++)
                nextGen[i] = new LineSegment();

            // One iteration step of Koch snowflake
            // For each line in generation
            // - divide the line into four
            //      - compute vector E(A,B)
            //      - S1 = E.A to E.A+(1/3)*E(A,B)
            //      - rotate the vector by 60°
            //      - next segment S2 from S1.B to S1.B+(1/3)*E(A,B)
            //      - rotate and create next segment etc.

            double sixtyDegrees = Math.PI / 3;

            //apply rule on each segment of the curve
            for (int i = 0; i < generation.Length; i++)
            {
                PointF lineDirection = new PointF((generation[i].B.X - generation[i].A.X),
                                                  (generation[i].B.Y - generation[i].A.Y));

                nextGen[i * 4 + 0].A = generation[i].A;
                nextGen[i * 4 + 0].B = new PointF(generation[i].A.X + lineDirection.X / 3f,
                                                  generation[i].A.Y + lineDirection.Y / 3f);

                lineDirection = RotateVector(lineDirection, sixtyDegrees);

                nextGen[i * 4 + 1].A = nextGen[i * 4 + 0].B;
                nextGen[i * 4 + 1].B = new PointF(nextGen[i * 4 + 0].B.X + lineDirection.X / 3f,
                                                  nextGen[i * 4 + 0].B.Y + lineDirection.Y / 3f);

                lineDirection = RotateVector(lineDirection, -2 * sixtyDegrees);

                nextGen[i * 4 + 2].A = nextGen[i * 4 + 1].B;
                nextGen[i * 4 + 2].B = new PointF(nextGen[i * 4 + 1].B.X + lineDirection.X / 3f,
                                                  nextGen[i * 4 + 1].B.Y + lineDirection.Y / 3f);

                nextGen[i * 4 + 3].A = nextGen[i * 4 + 2].B;
                nextGen[i * 4 + 3].B = generation[i].B;
            }

            return nextGen;
        }

        #endregion

#region Dragon
        internal LineSegment[] FirstGenerationDragon()
        {
            LineSegment[] square = new LineSegment[4];
            for (int i = 0; i < square.Length; i++)
                square[i] = new LineSegment();

            // Create the starting shape for dragon curve
            // - starting shape is a square
            //
            // Compute vector translation
            // - so the square is in the center of the canvas

            float lineLength = 0.5f * canvas.Height;

            square[0].A = new PointF(0, 0);
            square[0].B = new PointF(lineLength, 0);
            square[1].A = square[0].B;
            square[1].B = new PointF(lineLength, lineLength);
            square[2].A = square[1].B;
            square[2].B = new PointF(0, lineLength);
            square[3].A = square[2].B;
            square[3].B = square[0].A;

            // Compute translation vector
            float x = 0.5f * (canvas.Width - lineLength);
            float y = 0.5f * (canvas.Height - lineLength);
            translation = new PointF(x, y);

            return square;
        }

        private LineSegment[] NextDragonGeneration(LineSegment[] generation)
        {
            // Pocet hran je v dalsi generaci 2x vice
            nextGen = new LineSegment[generation.Length * 2];
            for (int i = 0; i < nextGen.Length; i++)
                nextGen[i] = new LineSegment();

            // Iteration step for dragon curve
            // - each line from original generation will be divided into 2
            //      - get vector of the line E(A,B)
            //      - rotate by (+-)45°
            //      - new line from E.A to E.A+(1/sqrt(2))E(A,B)
            //      - create the other segment
            // ATTENTION!! two consecutive lines from original generation must switch the "orientation" of their children
            // - one rotates by 45
            // - the next one rotates by -45

            double fortyFiveDegrees = Math.PI / 4;

            //apply rule on each segment of the curve
            for (int i = 0; i < generation.Length; i++)
            {
                PointF lineDirection = new PointF((generation[i].B.X - generation[i].A.X),
                                                  (generation[i].B.Y - generation[i].A.Y));

                //coef for switching direction of the generation
                float coef = 1;
                
                if (i % 2 == 0) coef = 1;
                else coef = -1;

                lineDirection = RotateVector(lineDirection, fortyFiveDegrees * coef);

                nextGen[i * 2 + 0].A = generation[i].A;
                nextGen[i * 2 + 0].B = new PointF(generation[i].A.X + lineDirection.X * (1.0f / (float)Math.Sqrt(2)),
                                                  generation[i].A.Y + lineDirection.Y * (1.0f / (float)Math.Sqrt(2)));

                nextGen[i * 2 + 1].A = nextGen[i * 2 + 0].B;
                nextGen[i * 2 + 1].B = generation[i].B;
            }

            return nextGen;
        }

#endregion

#region Special Koch

        /// <summary>
        /// Set up first generation of the special Koch snowflake
        /// </summary>
        /// <param name="g"></param>
        /// <param name="activeCanvas"></param>
        /// <returns></returns>
        internal LineSegment[] FirstGenerationSpecialKoch()
        {
            LineSegment[] triangle = new LineSegment[6];
            for (int i = 0; i < triangle.Length; i++)
                triangle[i] = new LineSegment();

            // Create an equilateral triangle
            // - same as normal Koch snowflake
            // - except add all the sides in both orientations
            // -> twice as many lines

            // Set up attribute translation
            // - value of translation to the center of the canvas
            // - centers the created triangle in the canvas

            LineSegment[] koch = FirstGeneration();
            for (int i = 0; i < koch.Length; i++)
            {
                triangle[i] = koch[i];
                triangle[koch.Length + i] = new LineSegment() { A = koch[i].B, B = koch[i].A };
            }

            return triangle;
        }

        private LineSegment[] NextSpecialKochGeneration(LineSegment[] generation)
        {
            // Next generation of special koch snowflake
            // - same rules as for normal snowflake, except save the created lines in both directions
            // -> twice as many lines

            LineSegment[] koch = NextKochGeneration(generation);

            // Copy in both directions
            nextGen = new LineSegment[(generation.Length * 4) * 2];
            for (int i = 0; i < koch.Length; i++)
            {
                nextGen[i] = koch[i];
                nextGen[i + koch.Length] =  new LineSegment() { A = koch[i].B, B = koch[i].A };
            }

            return nextGen;
        }
        #endregion

        /// <summary>
        /// Rotate a vector by a given angle
        /// </summary>
        /// <param name="vector"> Vector </param>
        /// <param name="angle"> Angle in radians </param>
        /// <returns> Rotated vector </returns>
        private PointF RotateVector(PointF vector, double angle)
        {
            vector = new PointF(vector.X * (float)Math.Cos(angle) - vector.Y * (float)Math.Sin(angle),
                                vector.X * (float)Math.Sin(angle) + vector.Y * (float)Math.Cos(angle));
            return vector;
        }

        /// <summary>
        /// Set curve type
        /// </summary>
        /// <param name="id"> New type </param>
        internal void Curve(int id)
        {
            this.curveID = id;
        }
    }
}